"""Defines functions converting raw instructions into textual form."""
from __future__ import print_function, absolute_import, division, unicode_literals

import itertools

from .opcodes import INSN_LEAVE_BLOCK, INSN_ENTER_BLOCK
from .decode import decode_bytecode
from .wasmtypes import VAL_TYPE_I32, VAL_TYPE_I64, VAL_TYPE_F32, VAL_TYPE_F64, MUTABLE, IMMUTABLE


def format_instruction(insn):
    """
    Takes a raw `Instruction` and translates it into a human readable text
    representation. As of writing, the text representation for WASM is not yet
    standardized, so we just emit some generic format.
    """
    text = insn.op.mnemonic

    if not insn.imm:
        return text

    return text + ' ' + ', '.join([
        getattr(insn.op.imm_struct, x.name).to_string(
            getattr(insn.imm, x.name)
        )
        for x in insn.op.imm_struct._meta.fields
    ])

_mutability_str_mapping = {
    MUTABLE: "mut",
    IMMUTABLE: ""
}

def format_mutability(mutability):
    """Takes a value type `int`, returning its string representation."""
    try:
        return _mutability_str_mapping[mutability]
    except KeyError:
        raise ValueError('Bad value for value type ({})'.format(mutability))

_lang_type_str_mapping = {
    VAL_TYPE_I32: 'i32',
    VAL_TYPE_I64: 'i64',
    VAL_TYPE_F32: 'f32',
    VAL_TYPE_F64: 'f64',
}


def format_lang_type(lang_type):
    """Takes a value type `int`, returning its string representation."""
    try:
        return _lang_type_str_mapping[lang_type]
    except KeyError:
        raise ValueError('Bad value for value type ({})'.format(lang_type))


def format_function(
    func_body,
    func_type=None,
    indent=2,
    format_locals=True,
):
    """
    Takes a `FunctionBody` and optionally a `FunctionType`, yielding the string 
    representation of the function line by line. The function type is required
    for formatting function parameter and return value information.
    """
    if func_type is None:
        yield 'func'
    else:
        param_section = ' (param {})'.format(' '.join(
            map(format_lang_type, func_type.param_types)
        )) if func_type.param_types else ''
        result_section = ' (result {})'.format(
            format_lang_type(func_type.return_type)
        ) if func_type.return_type else ''
        yield 'func' + param_section + result_section

    if format_locals and func_body.locals:
        yield '(locals {})'.format(' '.join(itertools.chain.from_iterable(
            itertools.repeat(format_lang_type(x.type), x.count)
            for x in func_body.locals
        )))

    level = 1
    for cur_insn in decode_bytecode(func_body.code):
        if cur_insn.op.flags & INSN_LEAVE_BLOCK:
            level -= 1
        yield ' ' * (level * indent) + format_instruction(cur_insn)
        if cur_insn.op.flags & INSN_ENTER_BLOCK:
            level += 1



"""Defines a simple, generic data (de)serialization mechanism."""
from __future__ import print_function, absolute_import, division, unicode_literals

from .compat import Callable, add_metaclass, byte2int, indent, deprecated_func
import collections
import logging
import struct as pystruct


logger = logging.getLogger()


class WasmField(object):
    """
    Abstract base class for all fields.

    Fields are purely a (de)serialization mechanism. They don't hold the value
    of decoded information, but take Python data-types and convert them
    to a raw byte format or vice versa. Thus, a field instance can be reused
    to de/encode multiple values.

    Besides the abstract interface, implements type counting and IDing to allow
    field order detection in Python 2, where `__prepare__` doesn't exist yet.
    In order to work correctly, field instances MUST NOT be shared between
    multiple structures using it but have to be instantiated per structure.
    """
    _type_ctr = 0

    def __init__(self):
        self._type_id = WasmField._type_ctr
        WasmField._type_ctr += 1

    def from_raw(self, struct, raw):
        raise NotImplementedError()

    def to_string(self, value):
        return repr(value)


class UIntNField(WasmField):
    """Field handling an unsigned LE int of fixed size."""
    CONVERTER_MAP = {
        8: pystruct.Struct('<B'),
        16: pystruct.Struct('<H'),
        32: pystruct.Struct('<I'),
        64: pystruct.Struct('<Q'),
    }

    def __init__(self, n, **kwargs):
        super(UIntNField, self).__init__(**kwargs)
        self.n = n
        self.byte_size = n // 8
        self.converter = self.CONVERTER_MAP[n]

    def from_raw(self, ctx, raw):
        return self.byte_size, self.converter.unpack(raw[:self.byte_size])[0], self

    def to_string(self, value):
        return hex(byte2int(value) if self.n == 8 else value)


class UnsignedLeb128Field(WasmField):
    """
    Field handling unsigned LEB128 values.
    https://en.wikipedia.org/wiki/LEB128
    """
    def from_raw(self, ctx, raw):
        offs = 0
        val = 0

        while True:
            segment = byte2int(raw[offs])
            val |= (segment & 0x7F) << (offs * 7)
            offs += 1
            if not (segment & 0x80):
                break

        return offs, val, self

    def to_string(self, value):
        return hex(value) if value > 1000 else str(value)


class SignedLeb128Field(WasmField):
    """
    Field handling signed LEB128 values.
    https://en.wikipedia.org/wiki/LEB128
    """
    def from_raw(self, ctx, raw):
        offs = 0
        val = 0
        bits = 0

        while True:
            segment = byte2int(raw[offs])
            val |= (segment & 0x7F) << bits
            offs += 1
            bits += 7
            if not (segment & 0x80):
                break

        if val & (1 << (bits - 1)):
            val -= 1 << bits

        return offs, val, self


class CondField(WasmField):
    """Optionalizes a field, depending on the context."""
    def __init__(self, field, condition, **kwargs):
        super(CondField, self).__init__(**kwargs)
        self.field = field
        self.condition = condition

    def from_raw(self, ctx, raw):
        if self.condition(ctx):
            return self.field.from_raw(ctx, raw)
        return 0, None, self

    def to_string(self, value):
        return 'None' if value is None else self.field.to_string(value)


class RepeatField(WasmField):
    """Repeats a field, having the repeat count depend on the context."""
    def __init__(self, field, repeat_count_getter, **kwargs):
        super(RepeatField, self).__init__(**kwargs)
        self.field = field
        self.repeat_count_getter = repeat_count_getter

    def from_raw(self, ctx, raw):
        repeat_count = self.repeat_count_getter(ctx)

        # Avoiding complex processing for byte arrays.
        if type(self.field) == UIntNField and self.field.n == 8:
            return repeat_count, raw[:repeat_count], self

        # For more complex types, invoke the field for parsing the
        # individual fields.
        offs = 0
        items = []
        for i in range(repeat_count):
            length, item, element_type = self.field.from_raw(ctx, raw[offs:])
            offs += length
            items.append(item)

        return offs, items, self

    def to_string(self, value):
        if value is None:
            return 'None'
        if len(value) > 100:
            return '<too long>'
        if len(value) == 0:
            return '[]'
        if isinstance(value[0], StructureData):
            return '\n' + indent(
                '\n'.join(self.field.to_string(x) for x in value),
                '  '
            )
        else:
            return '[' + ', '.join(self.field.to_string(x) for x in value) + ']'


class ConstField(WasmField):
    """Pseudo-Field, always returning a constant, consuming/generating no data."""
    def __init__(self, const, **kwargs):
        super(ConstField, self).__init__(**kwargs)
        self.const = const

    def from_raw(self, ctx, raw):
        return 0, self.const, self


class ChoiceField(WasmField):
    """Depending on context, either represent this or that field type."""
    _shared_none_field = ConstField(None)

    def __init__(self, choice_field_map, choice_getter, **kwargs):
        super(ChoiceField, self).__init__(**kwargs)
        self.choice_field_map = choice_field_map
        self.choice_getter = choice_getter

    def from_raw(self, ctx, raw):
        choice = self.choice_getter(ctx)
        if choice is None:
            return 0, None, self._shared_none_field
        return self.choice_field_map[choice].from_raw(ctx, raw)


class BytesField(RepeatField):
    """Shorthand for U8 `RepeatField`, adding string support."""
    def __init__(self, length_getter, is_str=False):
        super(BytesField, self).__init__(UIntNField(8), length_getter)
        self.is_str = is_str

    def to_string(self, value):
        if not self.is_str:
            return super(BytesField, self).to_string(value)

        try:
            return '"' + bytearray(value).decode('utf8') + '"'
        except UnicodeDecodeError:
            return '<bad utf8>'


FieldMeta = collections.namedtuple('FieldMeta', 'name field')


class MetaInfo(object):
    """Meta information for a `Structure`."""
    def __init__(self):
        self.fields = []
        self.data_class = None
        self.structure = None


class StructureData(object):
    """Base class for generated structure data classes."""
    __slots__ = ('_meta', '_decoder_meta')

    def __init__(self, for_decoding=False):
        self._decoder_meta = {'lengths': {}, 'types': {}} if for_decoding else None
        for cur_field_name, cur_field in self._meta.fields:
            setattr(self, cur_field_name, None)

    def get_meta(self):
        """
        Obtains meta info for this object. The object returned is shared
        between all objects of the same structure type.

        A getter is utilized here instead of a property to allow strict
        distinction of meta info from regular fields.
        """
        return self._meta

    def get_decoder_meta(self):
        """
        Obtains meta info from the decoder, like byte length in raw format.
        For objects not created through decoding, `None` is returned.
        """
        return self._decoder_meta

    @property
    @deprecated_func
    def _data_meta(self):
        """Property emulating old name of `_decoder_meta`. Deprecated, do not use."""
        return self._decoder_meta


class StructureMeta(type):
    """
    Metaclass used to create `Structure` classes,
    populating their `_meta` field and performing sanity checks.
    """
    def __new__(mcs, name, bases, cls_dict):
        # Inject _meta.
        meta = cls_dict['_meta'] = MetaInfo()

        # Iterate over fields, move relevant data to meta.
        for cur_field_name, cur_field in list(cls_dict.items()):
            # Is callable, property, private or magic? We don't touch those.
            if (
                isinstance(cur_field, Callable) or
                isinstance(cur_field, property) or
                cur_field_name.startswith('_')
            ):
                pass

            # Is one of our types? Metafy.
            elif isinstance(cur_field, WasmField):
                meta.fields.append(FieldMeta(cur_field_name, cur_field))

            # Unknown type, print warning.
            else:
                logger.warning(
                    'Non-WasmField typed field "{}" found on type "{}". '
                    'Ignoring.'.format(cur_field_name, name)
                )

        # Order fields by type ID (see `WasmField` for the "why").
        meta.fields = sorted(meta.fields, key=lambda x: x.field._type_id)

        # Create data class type for "instances".
        class GeneratedStructureData(StructureData):
            __slots__ = [x for x, _ in meta.fields]
            _meta = meta
        meta.data_class = GeneratedStructureData

        # Create class, saving type ref in meta.
        meta.structure = type.__new__(mcs, name, bases, cls_dict)
        return meta.structure


@add_metaclass(StructureMeta)
class Structure(WasmField):
    """Represents a collection of named fields."""
    def from_raw(self, ctx, raw):
        offs = 0
        data = self._meta.data_class(for_decoding=True)
        for cur_field_name, cur_field in self._meta.fields:
            data_len, val, data_type = cur_field.from_raw(data, raw[offs:])
            setattr(data, cur_field_name, val)
            decoder_meta = data.get_decoder_meta()
            decoder_meta['lengths'][cur_field_name] = data_len
            decoder_meta['types'][cur_field_name] = data_type
            offs += data_len
        return offs, data, self

    def to_string(self, value):
        lines = ['- [ {}'.format(self.__class__.__name__)]
        for cur_field_name, cur_field in self._meta.fields:
            field_val = getattr(value, cur_field_name)
            field_type = value.get_decoder_meta()['types'][cur_field_name]
            if isinstance(field_val, StructureData):
                lines.append('  | {} =\n{}'.format(
                    cur_field_name,
                    indent(field_type.to_string(field_val), '  ')
                ))
            else:
                lines.append('  | {} = {}'.format(
                    cur_field_name,
                    field_type.to_string(field_val)
                ))

        return '\n'.join(lines)



"""Provides functions for decoding WASM modules and bytecode."""
from __future__ import print_function, absolute_import, division, unicode_literals

from collections import namedtuple
from .modtypes import ModuleHeader, Section, SEC_UNK, SEC_NAME, NameSubSection
from .opcodes import OPCODE_MAP
from .compat import byte2int


Instruction = namedtuple('Instruction', 'op imm len')
ModuleFragment = namedtuple('ModuleFragment', 'type data')


def decode_bytecode(bytecode):
    """Decodes raw bytecode, yielding `Instruction`s."""
    bytecode_wnd = memoryview(bytecode)
    while bytecode_wnd:
        opcode_id = byte2int(bytecode_wnd[0])
        opcode = OPCODE_MAP[opcode_id]

        if opcode.imm_struct is not None:
            offs, imm, _ = opcode.imm_struct.from_raw(None, bytecode_wnd[1:])
        else:
            imm = None
            offs = 0

        insn_len = 1 + offs
        yield Instruction(opcode, imm, insn_len)
        bytecode_wnd = bytecode_wnd[insn_len:]


def decode_module(module, decode_name_subsections=False):
    """Decodes raw WASM modules, yielding `ModuleFragment`s."""
    module_wnd = memoryview(module)

    # Read & yield module header.
    hdr = ModuleHeader()
    hdr_len, hdr_data, _ = hdr.from_raw(None, module_wnd)
    yield ModuleFragment(hdr, hdr_data)
    module_wnd = module_wnd[hdr_len:]

    # Read & yield sections.
    while module_wnd:
        sec = Section()
        sec_len, sec_data, _ = sec.from_raw(None, module_wnd)

        # If requested, decode name subsections when encountered.
        if (
            decode_name_subsections and
            sec_data.id == SEC_UNK and
            sec_data.name == SEC_NAME
        ):
            sec_wnd = sec_data.payload
            while sec_wnd:
                subsec = NameSubSection()
                subsec_len, subsec_data, _ = subsec.from_raw(None, sec_wnd)
                yield ModuleFragment(subsec, subsec_data)
                sec_wnd = sec_wnd[subsec_len:]
        else:
            yield ModuleFragment(sec, sec_data)

        module_wnd = module_wnd[sec_len:]



"""Defines data structures used in WASM (binary) modules."""
from __future__ import print_function, absolute_import, division, unicode_literals

from .wasmtypes import *
from .opcodes import OP_END
from .types import (
    Structure, CondField, RepeatField,
    ChoiceField, WasmField, ConstField, BytesField,
)


class ModuleHeader(Structure):
    magic = UInt32Field()
    version = UInt32Field()


class FunctionImportEntryData(Structure):
    type = VarUInt32Field()


class ResizableLimits(Structure):
    flags = VarUInt32Field()
    initial = VarUInt32Field()
    maximum = CondField(VarUInt32Field(), lambda x: x.flags & 1)


class TableType(Structure):
    element_type = ElementTypeField()
    limits = ResizableLimits()


class MemoryType(Structure):
    limits = ResizableLimits()


class GlobalType(Structure):
    content_type = ValueTypeField()
    mutability = VarUInt1Field()


class ImportEntry(Structure):
    module_len = VarUInt32Field()
    module_str = BytesField(lambda x: x.module_len, is_str=True)
    field_len = VarUInt32Field()
    field_str = BytesField(lambda x: x.field_len, is_str=True)
    kind = ExternalKindField()
    type = ChoiceField({
        0: FunctionImportEntryData(),
        1: TableType(),
        2: MemoryType(),
        3: GlobalType(),
    }, lambda x: x.kind)


class ImportSection(Structure):
    count = VarUInt32Field()
    entries = RepeatField(ImportEntry(), lambda x: x.count)


class FuncType(Structure):
    form = VarInt7Field()
    param_count = VarUInt32Field()
    param_types = RepeatField(ValueTypeField(), lambda x: x.param_count)
    return_count = VarUInt1Field()
    return_type = CondField(ValueTypeField(), lambda x: bool(x.return_count))


class TypeSection(Structure):
    count = VarUInt32Field()
    entries = RepeatField(FuncType(), lambda x: x.count)


class FunctionSection(Structure):
    count = VarUInt32Field()
    types = RepeatField(VarUInt32Field(), lambda x: x.count)


class TableSection(Structure):
    count = VarUInt32Field()
    entries = RepeatField(TableType(), lambda x: x.count)


class MemorySection(Structure):
    count = VarUInt32Field()
    entries = RepeatField(MemoryType(), lambda x: x.count)


class InitExpr(WasmField):
    def from_raw(self, struct, raw):
        from .decode import decode_bytecode

        offs = 0
        instrs = []
        for cur_insn in decode_bytecode(raw):
            offs += cur_insn.len
            instrs.append(cur_insn)
            if cur_insn.op.id == OP_END:
                break

        return offs, instrs, self


class GlobalEntry(Structure):
    type = GlobalType()
    init = InitExpr()


class GlobalSection(Structure):
    count = VarUInt32Field()
    globals = RepeatField(GlobalEntry(), lambda x: x.count)


class ExportEntry(Structure):
    field_len = VarUInt32Field()
    field_str = BytesField(lambda x: x.field_len, is_str=True)
    kind = ExternalKindField()
    index = VarUInt32Field()


class ExportSection(Structure):
    count = VarUInt32Field()
    entries = RepeatField(ExportEntry(), lambda x: x.count)


class StartSection(Structure):
    index = VarUInt32Field()


class ElementSegment(Structure):
    index = VarUInt32Field()
    offset = InitExpr()
    num_elem = VarUInt32Field()
    elems = RepeatField(VarUInt32Field(), lambda x: x.num_elem)


class ElementSection(Structure):
    count = VarUInt32Field()
    entries = RepeatField(ElementSegment(), lambda x: x.count)


class LocalEntry(Structure):
    count = VarUInt32Field()
    type = ValueTypeField()


class FunctionBody(Structure):
    body_size = VarUInt32Field()
    local_count = VarUInt32Field()
    locals = RepeatField(
        LocalEntry(),
        lambda x: x.local_count,
    )
    code = BytesField(
        lambda x: (
            x.body_size -
            x.get_decoder_meta()['lengths']['local_count'] -
            x.get_decoder_meta()['lengths']['locals']
        )
    )


class CodeSection(Structure):
    count = VarUInt32Field()
    bodies = RepeatField(FunctionBody(), lambda x: x.count)


class DataSegment(Structure):
    index = VarUInt32Field()
    offset = InitExpr()
    size = VarUInt32Field()
    data = BytesField(lambda x: x.size)


class DataSection(Structure):
    count = VarUInt32Field()
    entries = RepeatField(DataSegment(), lambda x: x.count)


class DataCountSection(Structure):
    count = VarUInt32Field()


class Naming(Structure):
    index = VarUInt32Field()
    name_len = VarUInt32Field()
    name_str = BytesField(lambda x: x.name_len, is_str=True)


class NameMap(Structure):
    count = VarUInt32Field()
    names = RepeatField(Naming(), lambda x: x.count)


class LocalNames(Structure):
    index = VarUInt32Field()
    local_map = NameMap()


class LocalNameMap(Structure):
    count = VarUInt32Field()
    funcs = RepeatField(LocalNames, lambda x: x.count)


class NameSubSection(Structure):
    name_type = VarUInt7Field()
    payload_len = VarUInt32Field()
    payload = ChoiceField({
        NAME_SUBSEC_FUNCTION: NameMap(),
        NAME_SUBSEC_LOCAL: LocalNameMap(),
    }, lambda x: x.name_type)


class Section(Structure):
    id = VarUInt7Field()
    payload_len = VarUInt32Field()
    name_len = CondField(
        VarUInt32Field(),
        lambda x: x.id == 0,
    )
    name = CondField(
        BytesField(lambda x: x.name_len, is_str=True),
        lambda x: x.id == 0,
    )

    payload = ChoiceField({
        SEC_UNK: BytesField(lambda x: (
            x.payload_len -
            x.get_decoder_meta()['lengths']['name'] -
            x.get_decoder_meta()['lengths']['name_len']
        )),
        SEC_TYPE: TypeSection(),
        SEC_IMPORT: ImportSection(),
        SEC_FUNCTION: FunctionSection(),
        SEC_TABLE: TableSection(),
        SEC_MEMORY: MemorySection(),
        SEC_GLOBAL: GlobalSection(),
        SEC_EXPORT: ExportSection(),
        SEC_START: StartSection(),
        SEC_ELEMENT: ElementSection(),
        SEC_CODE: CodeSection(),
        SEC_DATA: DataSection(),
        SEC_DATA_COUNT: DataCountSection(),
    }, lambda x: x.id)

    overhang = BytesField(lambda x: max(0, (
        x.payload_len -
        x.get_decoder_meta()['lengths']['name'] -
        x.get_decoder_meta()['lengths']['name_len'] -
        x.get_decoder_meta()['lengths']['payload']
    )))



from __future__ import unicode_literals

__version__ = '1.0.0'

from .decode import (
    decode_bytecode,
    decode_module,
)

from .formatter import (
    format_function,
    format_instruction,
    format_lang_type,
    format_mutability,
)

from .modtypes import (
    ModuleHeader,
    FunctionImportEntryData,
    ResizableLimits,
    TableType,
    MemoryType,
    GlobalType,
    ImportEntry,
    ImportSection,
    FuncType,
    TypeSection,
    FunctionSection,
    TableSection,
    MemorySection,
    InitExpr,
    GlobalEntry,
    GlobalSection,
    ExportEntry,
    ExportSection,
    StartSection,
    ElementSegment,
    ElementSection,
    LocalEntry,
    FunctionBody,
    CodeSection,
    DataSegment,
    DataSection,
    Naming,
    NameMap,
    LocalNames,
    LocalNameMap,
    NameSubSection,
    Section,
)

from .immtypes import (
    BlockImm,
    BranchImm,
    BranchTableImm,
    CallImm,
    CallIndirectImm,
    LocalVarXsImm,
    GlobalVarXsImm,
    MemoryImm,
    CurGrowMemImm,
    I32ConstImm,
    I64ConstImm,
    F32ConstImm,
    F64ConstImm,
)

from .opcodes import (
    Opcode,
    INSN_ENTER_BLOCK,
    INSN_LEAVE_BLOCK,
    INSN_BRANCH,
    INSN_NO_FLOW,
)

for cur_op in opcodes.OPCODES:
    globals()[
        'OP_' + cur_op.mnemonic.upper().replace('.', '_').replace('/', '_')
    ] = cur_op.id

from .wasmtypes import (
    UInt8Field,
    UInt16Field,
    UInt32Field,
    UInt64Field,
    VarUInt1Field,
    VarUInt7Field,
    VarUInt32Field,
    VarInt7Field,
    VarInt32Field,
    VarInt64Field,
    ElementTypeField,
    ValueTypeField,
    ExternalKindField,
    BlockTypeField,
    SEC_UNK,
    SEC_TYPE,
    SEC_IMPORT,
    SEC_FUNCTION,
    SEC_TABLE,
    SEC_MEMORY,
    SEC_GLOBAL,
    SEC_EXPORT,
    SEC_START,
    SEC_ELEMENT,
    SEC_CODE,
    SEC_DATA,
    SEC_DATA_COUNT,
    SEC_NAME,
    LANG_TYPE_I32,
    LANG_TYPE_I64,
    LANG_TYPE_F32,
    LANG_TYPE_F64,
    LANG_TYPE_ANYFUNC,
    LANG_TYPE_FUNC,
    LANG_TYPE_EMPTY,
    VAL_TYPE_I32,
    VAL_TYPE_I64,
    VAL_TYPE_F32,
    VAL_TYPE_F64,
    NAME_SUBSEC_FUNCTION,
    NAME_SUBSEC_LOCAL,
    IMMUTABLE,
    MUTABLE,
)



"""Defines compatibility quirks for Python 2.7."""
from __future__ import print_function, absolute_import, division, unicode_literals

import collections
import sys
import functools
import logging
import warnings


def add_metaclass(metaclass):
    """
    Class decorator for creating a class with a metaclass.
    Borrowed from `six` module.
    """
    @functools.wraps(metaclass)
    def wrapper(cls):
        orig_vars = cls.__dict__.copy()
        slots = orig_vars.get('__slots__')
        if slots is not None:
            if isinstance(slots, str):
                slots = [slots]
            for slots_var in slots:
                orig_vars.pop(slots_var)
        orig_vars.pop('__dict__', None)
        orig_vars.pop('__weakref__', None)
        return metaclass(cls.__name__, cls.__bases__, orig_vars)
    return wrapper


def indent(text, prefix, predicate=None):
    """Adds 'prefix' to the beginning of selected lines in 'text'.

    If 'predicate' is provided, 'prefix' will only be added to the lines
    where 'predicate(line)' is True. If 'predicate' is not provided,
    it will default to adding 'prefix' to all non-empty lines that do not
    consist solely of whitespace characters.

    Borrowed from Py3 `textwrap` module.
    """
    if predicate is None:
        def predicate(line):
            return line.strip()

    def prefixed_lines():
        for line in text.splitlines(True):
            yield (prefix + line if predicate(line) else line)
    return ''.join(prefixed_lines())


def deprecated_func(func):
    """Deprecates a function, printing a warning on the first usage."""

    # We use a mutable container here to work around Py2's lack of
    # the `nonlocal` keyword.
    first_usage = [True]

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if first_usage[0]:
            warnings.warn(
                "Call to deprecated function {}.".format(func.__name__),
                DeprecationWarning,
            )
            first_usage[0] = False
        return func(*args, **kwargs)

    return wrapper


if sys.version_info[0] >= 3:
    def byte2int(x):
        return x

elif sys.version_info[0] == 2:
    def byte2int(x):
        return ord(x) if type(x) == str else x

else:
    raise Exception("Unsupported Python version")

if sys.version_info[0] == 3 and sys.version_info[1] < 10:
    Callable = collections.Callable
else:
    Callable = collections.abc.Callable



"""Defines immediate types for WASM bytecode instructions."""
from __future__ import print_function, absolute_import, division, unicode_literals

from .wasmtypes import *
from .types import Structure, RepeatField


class BlockImm(Structure):
    sig = BlockTypeField()


class BranchImm(Structure):
    relative_depth = VarUInt32Field()


class BranchTableImm(Structure):
    target_count = VarUInt32Field()
    target_table = RepeatField(VarUInt32Field(), lambda x: x.target_count)
    default_target = VarUInt32Field()


class CallImm(Structure):
    function_index = VarUInt32Field()


class CallIndirectImm(Structure):
    type_index = VarUInt32Field()
    reserved = VarUInt1Field()


class LocalVarXsImm(Structure):
    local_index = VarUInt32Field()


class GlobalVarXsImm(Structure):
    global_index = VarUInt32Field()


class MemoryImm(Structure):
    flags = VarUInt32Field()
    offset = VarUInt32Field()


class CurGrowMemImm(Structure):
    reserved = VarUInt1Field()


class I32ConstImm(Structure):
    value = VarInt32Field()


class I64ConstImm(Structure):
    value = VarInt64Field()


class F32ConstImm(Structure):
    value = UInt32Field()


class F64ConstImm(Structure):
    value = UInt64Field()



"""Defines types used for both modules and bytecode."""
from __future__ import print_function, absolute_import, division, unicode_literals

from .types import UIntNField, UnsignedLeb128Field, SignedLeb128Field


def _make_shortcut(klass, *args, **kwargs):
    def proxy(**kwargs2):
        kwargs.update(kwargs2)
        return klass(*args, **kwargs)
    return proxy


UInt8Field = _make_shortcut(UIntNField, 8)
UInt16Field = _make_shortcut(UIntNField, 16)
UInt32Field = _make_shortcut(UIntNField, 32)
UInt64Field = _make_shortcut(UIntNField, 64)

VarUInt1Field = _make_shortcut(UnsignedLeb128Field)
VarUInt7Field = _make_shortcut(UnsignedLeb128Field)
VarUInt32Field = _make_shortcut(UnsignedLeb128Field)

VarInt7Field = _make_shortcut(SignedLeb128Field)
VarInt32Field = _make_shortcut(SignedLeb128Field)
VarInt64Field = _make_shortcut(SignedLeb128Field)

ElementTypeField = VarInt7Field
ValueTypeField = VarInt7Field
ExternalKindField = UInt8Field
BlockTypeField = VarInt7Field


#
# Constants
#


# Section types.
SEC_UNK = 0
SEC_TYPE = 1
SEC_IMPORT = 2
SEC_FUNCTION = 3
SEC_TABLE = 4
SEC_MEMORY = 5
SEC_GLOBAL = 6
SEC_EXPORT = 7
SEC_START = 8
SEC_ELEMENT = 9
SEC_CODE = 10
SEC_DATA = 11
SEC_DATA_COUNT = 12
SEC_NAME = b'name'

# Language types.
LANG_TYPE_I32 = -0x01
LANG_TYPE_I64 = -0x02
LANG_TYPE_F32 = -0x03
LANG_TYPE_F64 = -0x04
LANG_TYPE_ANYFUNC = -0x10
LANG_TYPE_FUNC = -0x20
LANG_TYPE_EMPTY = -0x40

# Value types.
VAL_TYPE_I32 = LANG_TYPE_I32
VAL_TYPE_I64 = LANG_TYPE_I64
VAL_TYPE_F32 = LANG_TYPE_F32
VAL_TYPE_F64 = LANG_TYPE_F64

# Name subsection types.
NAME_SUBSEC_FUNCTION = 1
NAME_SUBSEC_LOCAL = 2

# Mutability in global types.
IMMUTABLE = 0
MUTABLE = 1



"""Testing & debug stuff."""
from __future__ import print_function, absolute_import, division, unicode_literals

import argparse
import sys

from .formatter import format_function
from .modtypes import SEC_CODE, SEC_TYPE, SEC_FUNCTION, Section
from .decode import decode_module


def dump():
    parser = argparse.ArgumentParser()
    parser.add_argument('wasm_file', type=str)
    parser.add_argument(
        '--disas',
        action='store_true',
        help="Disassemble code",
    )
    parser.add_argument(
        '--decode-names',
        action='store_true',
        help="Decode name section, if it exists",
    )
    args = parser.parse_args()

    try:
        with open(args.wasm_file, 'rb') as raw:
            raw = raw.read()
    except IOError as exc:
        print("[-] Can't open input file: " + str(exc), file=sys.stderr)
        return

    # Parse & print header.
    mod_iter = iter(decode_module(
        raw,
        decode_name_subsections=args.decode_names,
    ))
    hdr, hdr_data = next(mod_iter)
    print(hdr.to_string(hdr_data))

    # Parse & print other sections.
    code_sec = None
    type_sec = None
    func_sec = None
    for cur_sec, cur_sec_data in mod_iter:
        print(cur_sec.to_string(cur_sec_data))
        if type(cur_sec) == Section:
            if cur_sec_data.id == SEC_CODE:
                code_sec = cur_sec_data.payload
            elif cur_sec_data.id == SEC_TYPE:
                type_sec = cur_sec_data.payload
            elif cur_sec_data.id == SEC_FUNCTION:
                func_sec = cur_sec_data.payload

    # If ordered to disassemble, do so.
    # TODO: We might want to make use of debug names, if available.
    if args.disas and code_sec is not None:
        for i, func_body in enumerate(code_sec.bodies):
            print('{x} sub_{id:04X} {x}'.format(x='=' * 35, id=i))

            # If we have type info, use it.
            func_type = type_sec.entries[func_sec.types[i]] if (
                None not in (type_sec, func_sec)
            ) else None

            print()
            print('\n'.join(format_function(func_body, func_type)))
            print()



"""Defines mappings of opcodes to their info structures."""
from __future__ import print_function, absolute_import, division, unicode_literals

from collections import namedtuple
from .immtypes import *


Opcode = namedtuple('Opcode', 'id mnemonic imm_struct flags')

# Flags describing generic characteristics of instructions
INSN_ENTER_BLOCK = 1 << 0
INSN_LEAVE_BLOCK = 1 << 1
INSN_BRANCH = 1 << 2
INSN_NO_FLOW = 1 << 3  # does not pass control to next insn


OPCODES = [
    Opcode(0x00, 'unreachable',           None,                     INSN_NO_FLOW),
    Opcode(0x01, 'nop',                   None,                     0),
    Opcode(0x02, 'block',                 BlockImm(),               INSN_ENTER_BLOCK),
    Opcode(0x03, 'loop',                  BlockImm(),               INSN_ENTER_BLOCK),
    Opcode(0x04, 'if',                    BlockImm(),               INSN_ENTER_BLOCK),
    Opcode(0x05, 'else',                  None,                     INSN_ENTER_BLOCK | INSN_LEAVE_BLOCK),
    Opcode(0x0b, 'end',                   None,                     INSN_LEAVE_BLOCK),
    Opcode(0x0c, 'br',                    BranchImm(),              INSN_BRANCH),
    Opcode(0x0d, 'br_if',                 BranchImm(),              INSN_BRANCH),
    Opcode(0x0e, 'br_table',              BranchTableImm(),         INSN_BRANCH),
    Opcode(0x0f, 'return',                None,                     INSN_NO_FLOW),

    Opcode(0x10, 'call',                  CallImm(),                INSN_BRANCH),
    Opcode(0x11, 'call_indirect',         CallIndirectImm(),        INSN_BRANCH),

    Opcode(0x1a, 'drop',                  None,                     0),
    Opcode(0x1b, 'select',                None,                     0),

    Opcode(0x20, 'get_local',             LocalVarXsImm(),          0),
    Opcode(0x21, 'set_local',             LocalVarXsImm(),          0),
    Opcode(0x22, 'tee_local',             LocalVarXsImm(),          0),
    Opcode(0x23, 'get_global',            GlobalVarXsImm(),         0),
    Opcode(0x24, 'set_global',            GlobalVarXsImm(),         0),

    Opcode(0x28, 'i32.load',              MemoryImm(),              0),
    Opcode(0x29, 'i64.load',              MemoryImm(),              0),
    Opcode(0x2a, 'f32.load',              MemoryImm(),              0),
    Opcode(0x2b, 'f64.load',              MemoryImm(),              0),
    Opcode(0x2c, 'i32.load8_s',           MemoryImm(),              0),
    Opcode(0x2d, 'i32.load8_u',           MemoryImm(),              0),
    Opcode(0x2e, 'i32.load16_s',          MemoryImm(),              0),
    Opcode(0x2f, 'i32.load16_u',          MemoryImm(),              0),
    Opcode(0x30, 'i64.load8_s',           MemoryImm(),              0),
    Opcode(0x31, 'i64.load8_u',           MemoryImm(),              0),
    Opcode(0x32, 'i64.load16_s',          MemoryImm(),              0),
    Opcode(0x33, 'i64.load16_u',          MemoryImm(),              0),
    Opcode(0x34, 'i64.load32_s',          MemoryImm(),              0),
    Opcode(0x35, 'i64.load32_u',          MemoryImm(),              0),
    Opcode(0x36, 'i32.store',             MemoryImm(),              0),
    Opcode(0x37, 'i64.store',             MemoryImm(),              0),
    Opcode(0x38, 'f32.store',             MemoryImm(),              0),
    Opcode(0x39, 'f64.store',             MemoryImm(),              0),
    Opcode(0x3a, 'i32.store8',            MemoryImm(),              0),
    Opcode(0x3b, 'i32.store16',           MemoryImm(),              0),
    Opcode(0x3c, 'i64.store8',            MemoryImm(),              0),
    Opcode(0x3d, 'i64.store16',           MemoryImm(),              0),
    Opcode(0x3e, 'i64.store32',           MemoryImm(),              0),
    Opcode(0x3f, 'current_memory',        CurGrowMemImm(),          0),
    Opcode(0x40, 'grow_memory',           CurGrowMemImm(),          0),

    Opcode(0x41, 'i32.const',             I32ConstImm(),            0),
    Opcode(0x42, 'i64.const',             I64ConstImm(),            0),
    Opcode(0x43, 'f32.const',             F32ConstImm(),            0),
    Opcode(0x44, 'f64.const',             F64ConstImm(),            0),

    Opcode(0x45, 'i32.eqz',               None,                     0),
    Opcode(0x46, 'i32.eq',                None,                     0),
    Opcode(0x47, 'i32.ne',                None,                     0),
    Opcode(0x48, 'i32.lt_s',              None,                     0),
    Opcode(0x49, 'i32.lt_u',              None,                     0),
    Opcode(0x4a, 'i32.gt_s',              None,                     0),
    Opcode(0x4b, 'i32.gt_u',              None,                     0),
    Opcode(0x4c, 'i32.le_s',              None,                     0),
    Opcode(0x4d, 'i32.le_u',              None,                     0),
    Opcode(0x4e, 'i32.ge_s',              None,                     0),
    Opcode(0x4f, 'i32.ge_u',              None,                     0),
    Opcode(0x50, 'i64.eqz',               None,                     0),
    Opcode(0x51, 'i64.eq',                None,                     0),
    Opcode(0x52, 'i64.ne',                None,                     0),
    Opcode(0x53, 'i64.lt_s',              None,                     0),
    Opcode(0x54, 'i64.lt_u',              None,                     0),
    Opcode(0x55, 'i64.gt_s',              None,                     0),
    Opcode(0x56, 'i64.gt_u',              None,                     0),
    Opcode(0x57, 'i64.le_s',              None,                     0),
    Opcode(0x58, 'i64.le_u',              None,                     0),
    Opcode(0x59, 'i64.ge_s',              None,                     0),
    Opcode(0x5a, 'i64.ge_u',              None,                     0),
    Opcode(0x5b, 'f32.eq',                None,                     0),
    Opcode(0x5c, 'f32.ne',                None,                     0),
    Opcode(0x5d, 'f32.lt',                None,                     0),
    Opcode(0x5e, 'f32.gt',                None,                     0),
    Opcode(0x5f, 'f32.le',                None,                     0),
    Opcode(0x60, 'f32.ge',                None,                     0),
    Opcode(0x61, 'f64.eq',                None,                     0),
    Opcode(0x62, 'f64.ne',                None,                     0),
    Opcode(0x63, 'f64.lt',                None,                     0),
    Opcode(0x64, 'f64.gt',                None,                     0),
    Opcode(0x65, 'f64.le',                None,                     0),
    Opcode(0x66, 'f64.ge',                None,                     0),

    Opcode(0x67, 'i32.clz',               None,                     0),
    Opcode(0x68, 'i32.ctz',               None,                     0),
    Opcode(0x69, 'i32.popcnt',            None,                     0),
    Opcode(0x6a, 'i32.add',               None,                     0),
    Opcode(0x6b, 'i32.sub',               None,                     0),
    Opcode(0x6c, 'i32.mul',               None,                     0),
    Opcode(0x6d, 'i32.div_s',             None,                     0),
    Opcode(0x6e, 'i32.div_u',             None,                     0),
    Opcode(0x6f, 'i32.rem_s',             None,                     0),
    Opcode(0x70, 'i32.rem_u',             None,                     0),
    Opcode(0x71, 'i32.and',               None,                     0),
    Opcode(0x72, 'i32.or',                None,                     0),
    Opcode(0x73, 'i32.xor',               None,                     0),
    Opcode(0x74, 'i32.shl',               None,                     0),
    Opcode(0x75, 'i32.shr_s',             None,                     0),
    Opcode(0x76, 'i32.shr_u',             None,                     0),
    Opcode(0x77, 'i32.rotl',              None,                     0),
    Opcode(0x78, 'i32.rotr',              None,                     0),
    Opcode(0x79, 'i64.clz',               None,                     0),
    Opcode(0x7a, 'i64.ctz',               None,                     0),
    Opcode(0x7b, 'i64.popcnt',            None,                     0),
    Opcode(0x7c, 'i64.add',               None,                     0),
    Opcode(0x7d, 'i64.sub',               None,                     0),
    Opcode(0x7e, 'i64.mul',               None,                     0),
    Opcode(0x7f, 'i64.div_s',             None,                     0),
    Opcode(0x80, 'i64.div_u',             None,                     0),
    Opcode(0x81, 'i64.rem_s',             None,                     0),
    Opcode(0x82, 'i64.rem_u',             None,                     0),
    Opcode(0x83, 'i64.and',               None,                     0),
    Opcode(0x84, 'i64.or',                None,                     0),
    Opcode(0x85, 'i64.xor',               None,                     0),
    Opcode(0x86, 'i64.shl',               None,                     0),
    Opcode(0x87, 'i64.shr_s',             None,                     0),
    Opcode(0x88, 'i64.shr_u',             None,                     0),
    Opcode(0x89, 'i64.rotl',              None,                     0),
    Opcode(0x8a, 'i64.rotr',              None,                     0),
    Opcode(0x8b, 'f32.abs',               None,                     0),
    Opcode(0x8c, 'f32.neg',               None,                     0),
    Opcode(0x8d, 'f32.ceil',              None,                     0),
    Opcode(0x8e, 'f32.floor',             None,                     0),
    Opcode(0x8f, 'f32.trunc',             None,                     0),
    Opcode(0x90, 'f32.nearest',           None,                     0),
    Opcode(0x91, 'f32.sqrt',              None,                     0),
    Opcode(0x92, 'f32.add',               None,                     0),
    Opcode(0x93, 'f32.sub',               None,                     0),
    Opcode(0x94, 'f32.mul',               None,                     0),
    Opcode(0x95, 'f32.div',               None,                     0),
    Opcode(0x96, 'f32.min',               None,                     0),
    Opcode(0x97, 'f32.max',               None,                     0),
    Opcode(0x98, 'f32.copysign',          None,                     0),
    Opcode(0x99, 'f64.abs',               None,                     0),
    Opcode(0x9a, 'f64.neg',               None,                     0),
    Opcode(0x9b, 'f64.ceil',              None,                     0),
    Opcode(0x9c, 'f64.floor',             None,                     0),
    Opcode(0x9d, 'f64.trunc',             None,                     0),
    Opcode(0x9e, 'f64.nearest',           None,                     0),
    Opcode(0x9f, 'f64.sqrt',              None,                     0),
    Opcode(0xa0, 'f64.add',               None,                     0),
    Opcode(0xa1, 'f64.sub',               None,                     0),
    Opcode(0xa2, 'f64.mul',               None,                     0),
    Opcode(0xa3, 'f64.div',               None,                     0),
    Opcode(0xa4, 'f64.min',               None,                     0),
    Opcode(0xa5, 'f64.max',               None,                     0),
    Opcode(0xa6, 'f64.copysign',          None,                     0),

    Opcode(0xa7, 'i32.wrap/i64',          None,                     0),
    Opcode(0xa8, 'i32.trunc_s/f32',       None,                     0),
    Opcode(0xa9, 'i32.trunc_u/f32',       None,                     0),
    Opcode(0xaa, 'i32.trunc_s/f64',       None,                     0),
    Opcode(0xab, 'i32.trunc_u/f64',       None,                     0),
    Opcode(0xac, 'i64.extend_s/i32',      None,                     0),
    Opcode(0xad, 'i64.extend_u/i32',      None,                     0),
    Opcode(0xae, 'i64.trunc_s/f32',       None,                     0),
    Opcode(0xaf, 'i64.trunc_u/f32',       None,                     0),
    Opcode(0xb0, 'i64.trunc_s/f64',       None,                     0),
    Opcode(0xb1, 'i64.trunc_u/f64',       None,                     0),
    Opcode(0xb2, 'f32.convert_s/i32',     None,                     0),
    Opcode(0xb3, 'f32.convert_u/i32',     None,                     0),
    Opcode(0xb4, 'f32.convert_s/i64',     None,                     0),
    Opcode(0xb5, 'f32.convert_u/i64',     None,                     0),
    Opcode(0xb6, 'f32.demote/f64',        None,                     0),
    Opcode(0xb7, 'f64.convert_s/i32',     None,                     0),
    Opcode(0xb8, 'f64.convert_u/i32',     None,                     0),
    Opcode(0xb9, 'f64.convert_s/i64',     None,                     0),
    Opcode(0xba, 'f64.convert_u/i64',     None,                     0),
    Opcode(0xbb, 'f64.promote/f32',       None,                     0),

    Opcode(0xbc, 'i32.reinterpret/f32',   None,                     0),
    Opcode(0xbd, 'i64.reinterpret/f64',   None,                     0),
    Opcode(0xbe, 'f32.reinterpret/i32',   None,                     0),
    Opcode(0xbf, 'f64.reinterpret/i64',   None,                     0),
]

OPCODE_MAP = {x.id: x for x in OPCODES}

# Generate integer constants for opcodes.
for cur_op in OPCODES:
    globals()[
        'OP_' + cur_op.mnemonic.upper().replace('.', '_').replace('/', '_')
    ] = cur_op.id
